#
/*
 *	    M U L T I - D E V I C E   Q U E U E I N G   S Y S T E M
 *
 *			   Douglas P. Kingston III
 *
 *	       Imagen Laser Printer Ethernet Server (imeserver)
 *
 *	Note that a pipe to the Deamon will be on file descriptor 3.
 *	________________________________________________________________
 *
 *				Revision History
 *				================
 */

#include <stdio.h>
#include <signal.h>
#include <errno.h>
#include "queue.h"
#include "qreplys.h"

#if BSD42 | BSD41c
#include <sys/file.h>
#endif

char	*language;
char	*printhost;
int	printport;
struct	request r;
int	acctfd;
int	lpoptions;
int	(*process)();
char	*workdir;
char	*title;
char	*user;

char	inputfile[PATHSIZ];		/* Name of datafile with cmds */

extern	char *Qdatadir;

extern	char	*index();
extern	char	*malloc();
extern	char	*strdup();
extern	struct	request *gethdr();
extern	char	*xmalloc();
extern	char	*fixstr();
extern	int	impress();
extern	int	tektronix();
extern	int	printer();
extern	int	daisy();
int	dorestart();
int	doflush();

/*
 *		M A I N  ( )
 *  As a MDQS queue server, this process will be started up
 *  with the queue control file as the standard input, and
 *  the device of interest on the standard output.
 */
main (argc, argv)
int	argc;
char	**argv;
{
	char	linebuf[LINESIZ];
	char	*cp;
	int	fin;
	
	signal( SIGHUP, SIG_IGN );		/* Paranoid. */
	signal( SIGEMT, dorestart );
	signal( SIGTERM, doflush );

	user = strdup ("UnknownUser");
	title = strdup ("UnknownTitle");

	arginit (argc, argv);
	setuid(getuid());		/* return to real uid */
	
	if (gethdr (stdin, &r) == (struct request *) NULL) {
		fprintf (stderr, "IMSERVER: Error reading request header\n");
		retmsg (RP_FATAL, "Error reading request header");
		exit (10);
	}

	while (fgets (linebuf, sizeof (linebuf), stdin) != NULL) {
		if ((cp = index (linebuf, '\n')) != 0)
			*cp = 0;
		switch (linebuf[0]) {
		case 'D':
			workdir = strdup(&linebuf[1]);
			break;

		case 'F':
			if (workdir == 0)
				workdir = "/tmp";
			if (chdir (workdir) < 0)
				chdir ("/tmp");
			fin = open( &linebuf[1], READ );
			if( fin < 0 ) {
				fprintf( stderr, "IMSERVER: Can't open input file '%s'\n", &linebuf[1] );
				break;
			} else {
				sendit(fin);
				close (fin);
			}
			if (lpoptions & NOKEEP)
				unlink (&linebuf[1]);
			break;

		case 'I':
			sprintf( inputfile, "%s/%s", Qdatadir, &linebuf[1] );
			fin = open( inputfile, READ );
			if( fin < 0 ) {
				fprintf( stderr, "IMSERVER: Can't open input file '%s'\n", inputfile );
				retmsg (RP_FATAL, "Cannot open input file");
				exit( 99 );
			} else {
				sendit(fin);
				close (fin);
			}
			unlink( inputfile );
			break;

		case 'O':
			lpoptions = atoi(&linebuf[0]);
			break;

		case 'T':
			title = strdup(&linebuf[1]);
			break;

		case 'U':
			user = strdup(&linebuf[1]);
			break;

		default:
			/* Ignore unknown control lines */
			break;
		}
	}
	retmsg (RP_OK, 0);
	exit (0);
}

/*
 *  S E N D I T (  )
 *
 *  Send an input file to the imagen.
 */
sendit(fin)
register int fin;
{
	int	status;
	int	pagecnt;
	FILE	*ifp, *ofp;
	int	imfd;
	int	i;

	/*
	 *	Now we are ready to start the imagen transfer process.
	 *	We are already running with our UID/GID
	 *	the same as the invoker.
	 */
	if(( imfd = iconnect (printhost, printport)) < 0 ) {
		retmsg (RP_RESTART, "connect to %s/%d failed",
			printhost, printport);
		exit(0);
	}

	ofp = fdopen (imfd, "w");
	ifp = fdopen (fin, "r");
	doheader (ofp);
	pagecnt = (*process)(ifp, ofp);
	status = (ferror (ofp) <<8) | ferror (ifp);
	fclose (ofp);
	account(pagecnt, language, status, user);
}

doheader (fout)
FILE	*fout;
{
	register int c;
	char	*cp;
	time_t timenow;

	time (&timenow);
	fputs ("@Document(", fout);
	cp = fixstr(user);
	fprintf (fout, "Owner %s", cp);
	free (cp);
	cp = fixstr(title);
	fprintf (fout, ", Name %s", cp);
	free (cp);
	cp = fixstr(ctime(&r.r_tsubmit));
	fprintf (fout, ", Submitted %s", cp);
	free (cp);
	cp = fixstr(ctime (&r.r_tstart));
	fprintf (fout, ", Started %s", cp);
	free (cp);
	cp = fixstr(ctime (&timenow));
	fprintf (fout, ", Printed %s", cp);
	free (cp);
	fprintf (fout, ", Language %s", language);
	if (strcmp(language, "impress"))
		fputs (", Pagereversal", fout);
	fputs (", JobHeader, JamResistance)", fout);

	fflush(fout);
	if (ferror (fout)) {
		retmsg (RP_FATAL, "Error writing to imagen");
		exit (0);
	}
}

/*
 *		X M A L L O C ( )
 *	Just malloc(), but it always returns a valid pointer or exits
 */
char *
xmalloc( size )
int	size;
{
	register char	*cp;

	if(( cp = malloc( size )) == NULL ) {
		fprintf( stderr, "IMSERVER: No free memory!\n");
		retmsg(RP_FATAL, "No free memory");
		if( *inputfile )
			unlink( inputfile );
		exit( 0 );
	}
	return( cp );
}

/*
 *	DORESTART ()
 *
 *  This function is called by a signal to restart a job.
 */
dorestart()
{
	retmsg (RP_RESTART, "RESTART requested");
	exit( 0 );
}

/*
 *	DOFLUSH ()
 *
 *  This function is called by a signal to "flush" the job.
 */
doflush()
{
	retmsg (RP_FLUSH, "Job flushed on request");
	exit( 0 );
}

/*
 *	A R G I N I T
 *
 *	Parse args of lineprinter like servers
 */
arginit(argc, argv)
int	argc;
char	**argv;
{
	register int i;

	if(argc < 4) {
		retmsg (RP_FATAL, "IMESERVER called improperly");
		fprintf (stderr, "IMESERVER: Missing language argument");
		exit (1);
	}
	if (strcmp (argv[1], "impress") == 0)
		process = impress;
	else if (strcmp (argv[1], "tektronix") == 0)
		process = tektronix;
	else if (strcmp (argv[1], "printer") == 0)
		process = printer;
	else if (strcmp (argv[1], "daisy") == 0)
		process = daisy;
	else {
		retmsg (RP_FATAL, "IMESERVER called with bad language");
		fprintf (stderr, "IMESERVER: Bad language argument");
		exit (1);
	}
	language = argv[1];


#if BSD42 | SYS3
	acctfd = open (argv[2], 1, O_APPEND);
#else
	acctfd = open (argv[2], 2);
#endif

/*		retmsg (RP_FATAL, "IMESERVER cannot open accounting file");
		fprintf (stderr, "IMESERVER: Cannot open accounting file");
		exit (1);
*/

	printhost = argv[3];
	if (argc > 4)
		printport = atoi(argv[4]);
}

/*
 *	F I X S T R  (  )
 *
 *  Sanitize strings for Imagen document control
 */
char *
fixstr (cp)
register char *cp;
{
	char	tmpbuf[300];
	register char	*tp;

	tp = tmpbuf;
	*tp++ = '"';
	while (*cp) {
		switch (*cp) {
		case '"':		/* Swallow */
			cp++;
			break;
		case '@':
			*tp++ = ' ';
			*tp++ = 'a'; *tp++ = 't';
			*tp++ = ' ';
			cp++;
			break;
		default:
			*tp++ = *cp++;
			continue;

		}
	}
	*tp++ = '"';
	*tp = 0;
	return (strdup (tmpbuf));
}

account (pagecnt, language, status, user)
int	pagecnt;
char	*language;
int	status;
char	*user;
{
	char	buf[512];
	int	len;
	time_t	timenow;

	time (&timenow);
	sprintf(buf, "%d %c %o/%o %12.12s %s\n", pagecnt, *language,
		(status>>8)&0377, status&0377, ctime(&timenow)+4, user);
	len = strlen(buf);
	if (acctfd != -1) {
		lseek (acctfd, 0L, 2);
		if (write (acctfd, buf, len) != len) {
			retmsg (RP_FATAL, "Imserver write to accounting file");
			fprintf (stderr, "IMESERVER: Cannot write to accounting file");
			exit (1);
		}
	}
}

/* Network portion follows */

#include <sys/socket.h>
#include <sys/file.h>

#include <netinet/in.h>
#include <netdb.h>

struct	sockaddr_in sin = { AF_INET };
int 	s;
int	options;
char	buf[BUFSIZ];

iconnect (printhost, printport)
char	*printhost;
int	printport;
{
	int on = 1;
	char *cp;
	struct hostent *host;
	struct servent *sp;

	host = gethostbyname(printhost);
	if (host) {
		sin.sin_family = host->h_addrtype;
		bcopy(host->h_addr, (caddr_t)&sin.sin_addr, host->h_length);
	} else {
		sin.sin_family = AF_INET;
		sin.sin_addr.s_addr = inet_addr(printhost);
		if (sin.sin_addr.s_addr == -1) {
			fprintf(stderr, "IMESERVER: %s: bad hostname/number\n", printhost);
			retmsg (RP_FATAL, "Unable to determine host address");
			exit (0);
		}
	}
	if (printport)
		sin.sin_port = htons(printport);
	else if (sp = getservbyname("iostation", "tcp"))
		sin.sin_port = sp->s_port;
	else {
		fprintf (stderr, "IMESERVER: port not specified or defaulted");
		retmsg (RP_FATAL, "printer port not specified to imeserver");
		exit (0);
	}

	s = socket(AF_INET, SOCK_STREAM, 0);
	if (s < 0)
		return (-1);
	if (setsockopt(s, SOL_SOCKET, SO_KEEPALIVE, &on, sizeof on) < 0) {
		fprintf (stderr, "bad setsockopt (SO_KEEPALIVE)");
		retmsg (RP_NOPEN, "bad setsockopt (SO_KEEPALIVE)");
		exit (0);
	}

	statmsg("attempting connection to laser printer");
	if (connect(s, &sin, sizeof (sin))) {
		sprintf (buf, "connect failed: %s", sys_errlist[errno]);
		retmsg (RP_NOPEN, buf);
		exit (0);
	}
	statmsg("connected to laser printer");
	return (s);
}
